@@ -58,9 +58,10 @@ SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #include "config.h"
 #endif
 
+#include <string.h>
+
 #include "xf86.h"
 #include "xf86_OSproc.h"
-#include "xf86_ansic.h"
 #include "compiler.h"
 
 #include "xf86fbman.h"
@@ -86,32 +87,53 @@ I830InitHWCursor(ScrnInfoPtr pScrn)
 
    DPRINTF(PFX, "I830InitHWCursor\n");
    /* Initialise the HW cursor registers, leaving the cursor hidden. */
-   if (IS_MOBILE(pI830) || IS_I915G(pI830) || IS_I945G(pI830)) {
+   if (IS_MOBILE(pI830) || IS_I9XX(pI830)) {
       temp = INREG(CURSOR_A_CONTROL);
       temp &= ~(CURSOR_MODE | MCURSOR_GAMMA_ENABLE | MCURSOR_MEM_TYPE_LOCAL |
 		MCURSOR_PIPE_SELECT);
       temp |= CURSOR_MODE_DISABLE;
       temp |= (pI830->pipe << 28);
+      if (pI830->CursorIsARGB)
+         temp |= CURSOR_MODE_64_ARGB_AX | MCURSOR_GAMMA_ENABLE;
+      else
+         temp |= CURSOR_MODE_64_4C_AX;
       /* Need to set control, then address. */
       OUTREG(CURSOR_A_CONTROL, temp);
-      if (pI830->CursorIsARGB)
-         OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
-      else
-         OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
-      if (pI830->Clone) {
+      if (pI830->CursorNeedsPhysical) {
+         if (pI830->CursorIsARGB)
+            OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
+         else
+            OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
+      } else {
+         if (pI830->CursorIsARGB)
+            OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Start);
+         else
+            OUTREG(CURSOR_A_BASE, pI830->CursorMem->Start);
+      }
+      if (pI830->Clone || pI830->MergedFB) {
          temp &= ~MCURSOR_PIPE_SELECT;
          temp |= (!pI830->pipe << 28);
          OUTREG(CURSOR_B_CONTROL, temp);
-         if (pI830->CursorIsARGB)
-            OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
-         else
-            OUTREG(CURSOR_B_BASE, pI830->CursorMem->Physical);
+         if (pI830->CursorNeedsPhysical) {
+            if (pI830->CursorIsARGB)
+               OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
+            else
+               OUTREG(CURSOR_B_BASE, pI830->CursorMem->Physical);
+	 } else {
+            if (pI830->CursorIsARGB)
+               OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Start);
+            else
+               OUTREG(CURSOR_B_BASE, pI830->CursorMem->Start);
+	 }
       }
    } else {
       temp = INREG(CURSOR_CONTROL);
       temp &= ~(CURSOR_FORMAT_MASK | CURSOR_GAMMA_ENABLE |
 		CURSOR_ENABLE  | CURSOR_STRIDE_MASK);
-      temp |= (CURSOR_FORMAT_3C);
+      if (pI830->CursorIsARGB)
+         temp |= CURSOR_FORMAT_ARGB | CURSOR_GAMMA_ENABLE;
+      else 
+         temp |= CURSOR_FORMAT_3C;
       /* This initialises the format and leave the cursor disabled. */
       OUTREG(CURSOR_CONTROL, temp);
       /* Need to set address and size after disabling. */
@@ -139,8 +161,8 @@ I830CursorInit(ScreenPtr pScreen)
    if (!infoPtr)
       return FALSE;
 
-   infoPtr->MaxWidth = 64;
-   infoPtr->MaxHeight = 64;
+   infoPtr->MaxWidth = I810_CURSOR_X;
+   infoPtr->MaxHeight = I810_CURSOR_Y;
    infoPtr->Flags = (HARDWARE_CURSOR_TRUECOLOR_AT_8BPP |
 		     HARDWARE_CURSOR_BIT_ORDER_MSBFIRST |
 		     HARDWARE_CURSOR_INVERT_MASK |
@@ -155,6 +177,8 @@ I830CursorInit(ScreenPtr pScreen)
    infoPtr->ShowCursor = I830ShowCursor;
    infoPtr->UseHWCursor = I830UseHWCursor;
 
+   pI830->pCurs = NULL;
+
 #ifdef ARGB_CURSOR
    pI830->CursorIsARGB = FALSE;
 
@@ -179,6 +203,8 @@ I830UseHWCursor(ScreenPtr pScreen, CursorPtr pCurs)
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    I830Ptr pI830 = I830PTR(pScrn);
 
+   pI830->pCurs = pCurs;
+
    DPRINTF(PFX, "I830UseHWCursor\n");
    if (pI830->CursorNeedsPhysical && !pI830->CursorMem->Physical) 
       return FALSE;
@@ -198,6 +224,52 @@ I830LoadCursorImage(ScrnInfoPtr pScrn, unsigned char *src)
 #ifdef ARGB_CURSOR
    pI830->CursorIsARGB = FALSE;
 #endif
+ 
+   memset(pcurs, 0, 64 * 64 / 4);
+
+#define GetBit(image, x, y)\
+    ((int)((*(image + ((x) / 8) + ((y) * (128/8))) &\
+	    (1 << ( 7 -((x) % 8) ))) ? 1 : 0))
+
+#define SetBit(image, x, y)\
+    (*(image + (x) / 8 + (y) * (128/8)) |=\
+     (int) (1 <<  (7-((x) % 8))))
+
+   switch (pI830->rotation) {
+      case RR_Rotate_90:
+         for (y = 0; y < 64; y++) {
+            for (x = 0; x < 64; x++) {
+               if (GetBit(src, 64 - y - 1, x))
+                  SetBit(pcurs, x, y);
+               if (GetBit(src, 128 - y - 1, x))
+                  SetBit(pcurs, x + 64, y);
+            }
+         }
+
+         return;
+      case RR_Rotate_180:
+         for (y = 0; y < 64; y++) {
+            for (x = 0; x < 64; x++) {
+               if (GetBit(src, 64 - x - 1, 64 - y - 1))
+                  SetBit(pcurs, x, y);
+               if (GetBit(src, 128 - x - 1, 64 - y - 1))
+                  SetBit(pcurs, x + 64, y);
+            }
+         }
+
+         return;
+      case RR_Rotate_270:
+         for (y = 0; y < 64; y++) {
+            for (x = 0; x < 64; x++) {
+               if (GetBit(src, y, 64 - x - 1))
+                  SetBit(pcurs, x, y);
+               if (GetBit(src, y + 64, 64 - x - 1))
+                  SetBit(pcurs, x + 64, y);
+            }
+         }
+
+         return;
+   }
 
    for (y = 0; y < 64; y++) {
       for (x = 0; x < 64 / 4; x++) {
@@ -215,6 +287,9 @@ static Bool I830UseHWCursorARGB (ScreenPtr pScreen, CursorPtr pCurs)
    I830Ptr pI830 = I830PTR(pScrn);
 
    DPRINTF(PFX, "I830UseHWCursorARGB\n");
+
+   pI830->pCurs = pCurs;
+
    if (pScrn->bitsPerPixel == 8)
       return FALSE;
 
@@ -239,13 +314,52 @@ static void I830LoadCursorARGB (ScrnInfoPtr pScrn, CursorPtr pCurs)
    if (!image)
 	return;	/* XXX can't happen */
     
-#ifdef ARGB_CURSOR
    pI830->CursorIsARGB = TRUE;
-#endif
 
    w = pCurs->bits->width;
    h = pCurs->bits->height;
 
+   switch (pI830->rotation) {
+      case RR_Rotate_90:
+         for (y = 0; y < h; y++) {
+            for (x = 0; x < w; x++)
+               dst[(y) + ((64 - x - 1) * 64)] = *image++;
+            for(; x < 64; x++)
+               dst[(y) + ((64 - x - 1) * 64)] = 0;
+         }
+         for(; y < 64; y++) {
+   	    for(x = 0; x < 64; x++)
+               dst[(y) + ((64 - x - 1) * 64)] = 0;
+         }
+         return;
+
+      case RR_Rotate_180:
+         for (y = 0; y < h; y++) {
+            for (x = 0; x < w; x++)
+               dst[(64 - x - 1) + ((64 - y - 1) * 64)] = *image++;
+            for(; x < 64; x++)
+               dst[(64 - x - 1) + ((64 - y - 1) * 64)] = 0;
+         }
+         for(; y < 64; y++) {
+            for(x = 0; x < 64; x++)
+               dst[(64 - x - 1) + ((64 - y - 1) * 64)] = 0;
+         }
+         return;
+
+      case RR_Rotate_270:
+         for (y = 0; y < h; y++) {
+            for (x = 0; x < w; x++)
+               dst[(64 - y - 1) + (x * 64)] = *image++;
+            for(; x < 64; x++)
+               dst[(64 - y - 1) + (x * 64)] = 0;
+         }
+         for(; y < 64; y++) {
+            for(x = 0; x < 64; x++)
+               dst[(64 - y - 1) + (x * 64)] = 0;
+         }
+         return;
+   }
+
    for(y = 0; y < h; y++) {
       for(x = 0; x < w; x++)
           *dst++ = *image++;
@@ -260,6 +374,102 @@ static void I830LoadCursorARGB (ScrnInfoPtr pScrn, CursorPtr pCurs)
 }
 #endif
 
+#define CDMPTR    ((I830ModePrivatePtr)pI830->currentMode->Private)->merged
+
+static void
+I830SetCursorPositionMerged(ScrnInfoPtr pScrn, int x, int y)
+{
+   I830Ptr pI830 = I830PTR(pScrn);
+   ScrnInfoPtr    pScrn2 = pI830->pScrn_2;
+   DisplayModePtr mode1 = CDMPTR.First;
+   Bool hide = FALSE, show = FALSE;
+   DisplayModePtr mode2 = CDMPTR.Second;
+   int x1, y1, x2, y2;
+   int total_y1 = pScrn->frameY1 - pScrn->frameY0;
+   int total_y2 = pScrn2->frameY1 - pScrn2->frameY0;
+   CARD32 temp = 0, temp2 = 0;
+
+   x += pScrn->frameX0;
+   y += pScrn->frameY0;
+
+   x1 = x - pI830->FirstframeX0;
+   y1 = y - pI830->FirstframeY0;
+
+   x2 = x - pScrn2->frameX0;
+   y2 = y - pScrn2->frameY0;
+
+   if (y1 > total_y1)
+      y1 = total_y1;
+   if (y2 > total_y2)                  
+      y2 = total_y2;
+
+   /* move cursor offscreen */
+   if (y1 >= 0 && y2 >= mode2->VDisplay) {
+      y2 = -I810_CURSOR_Y;  
+   } else if (y2 >= 0 && y1 >= mode1->VDisplay) {
+      y1 = -I810_CURSOR_Y;  
+   }
+   if (x1 >= 0 && x2 >= mode2->HDisplay) {
+      x2 = -I810_CURSOR_X;  
+   } else if (x2 >= 0 && x1 >= mode1->HDisplay) {
+      x1 = -I810_CURSOR_X;  
+   }
+
+   /* Clamp the cursor position to the visible screen area */
+   if (x1 >= mode1->HDisplay) x1 = mode1->HDisplay - 1;
+   if (y1 >= mode1->VDisplay) y1 = mode1->VDisplay - 1;
+   if (x1 <= -I810_CURSOR_X) x1 = -I810_CURSOR_X + 1;
+   if (y1 <= -I810_CURSOR_Y) y1 = -I810_CURSOR_Y + 1;
+   if (x2 >= mode2->HDisplay) x2 = mode2->HDisplay - 1;
+   if (y2 >= mode2->VDisplay) y2 = mode2->VDisplay - 1;
+   if (x2 <= -I810_CURSOR_X) x2 = -I810_CURSOR_X + 1;
+   if (y2 <= -I810_CURSOR_Y) y2 = -I810_CURSOR_Y + 1;
+
+   if (x1 < 0) {
+      temp |= (CURSOR_POS_SIGN << CURSOR_X_SHIFT);
+      x1 = -x1;
+   }
+   if (y1 < 0) {
+      temp |= (CURSOR_POS_SIGN << CURSOR_Y_SHIFT);
+      y1 = -y1;
+   }
+   if (x2 < 0) {
+      temp2 |= (CURSOR_POS_SIGN << CURSOR_X_SHIFT);
+      x2 = -x2;
+   }
+   if (y2 < 0) {
+      temp2 |= (CURSOR_POS_SIGN << CURSOR_Y_SHIFT);
+      y2 = -y2;
+   }
+
+   temp |= ((x1 & CURSOR_POS_MASK) << CURSOR_X_SHIFT);
+   temp |= ((y1 & CURSOR_POS_MASK) << CURSOR_Y_SHIFT);
+   temp2 |= ((x2 & CURSOR_POS_MASK) << CURSOR_X_SHIFT);
+   temp2 |= ((y2 & CURSOR_POS_MASK) << CURSOR_Y_SHIFT);
+
+   OUTREG(CURSOR_A_POSITION, temp);
+   OUTREG(CURSOR_B_POSITION, temp2);
+
+   if (pI830->cursorOn) {
+      if (hide)
+	 pI830->CursorInfoRec->HideCursor(pScrn);
+      else if (show)
+	 pI830->CursorInfoRec->ShowCursor(pScrn);
+      pI830->cursorOn = TRUE;
+   }
+
+   /* have to upload the base for the new position */
+   if (IS_I9XX(pI830)) {
+      if (pI830->CursorIsARGB) {
+         OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
+         OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
+      } else {
+         OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
+         OUTREG(CURSOR_B_BASE, pI830->CursorMem->Physical);
+      }
+   }
+}
+
 static void
 I830SetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
 {
@@ -267,7 +477,57 @@ I830SetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
    CARD32 temp = 0;
    static Bool outsideViewport = FALSE;
    Bool hide = FALSE, show = FALSE;
+   int oldx = x, oldy = y;
+   int hotspotx = 0, hotspoty = 0;
+#if 0
+   static Bool outsideViewport = FALSE;
+#endif
 
+   if (pI830->MergedFB) {
+      I830SetCursorPositionMerged(pScrn, x, y);
+      return;
+   }
+
+   oldx += pScrn->frameX0; /* undo what xf86HWCurs did */
+   oldy += pScrn->frameY0;
+
+   switch (pI830->rotation) {
+      case RR_Rotate_0:
+         x = oldx;
+         y = oldy;
+         break;
+      case RR_Rotate_90:
+         x = oldy;
+         y = pScrn->pScreen->width - oldx;
+         hotspoty = I810_CURSOR_X;
+         break;
+      case RR_Rotate_180:
+         x = pScrn->pScreen->width - oldx;
+         y = pScrn->pScreen->height - oldy;
+         hotspotx = I810_CURSOR_X;
+         hotspoty = I810_CURSOR_Y;
+         break;
+      case RR_Rotate_270:
+         x = pScrn->pScreen->height - oldy;
+         y = oldx;
+         hotspotx = I810_CURSOR_Y;
+         break;
+   }
+
+   x -= hotspotx;
+   y -= hotspoty;
+
+   /* Now, readjust */
+   x -= pScrn->frameX0;
+   y -= pScrn->frameY0;
+
+   /* Clamp the cursor position to the visible screen area */
+   if (x >= pScrn->currentMode->HDisplay) x = pScrn->currentMode->HDisplay - 1;
+   if (y >= pScrn->currentMode->VDisplay) y = pScrn->currentMode->VDisplay - 1;
+   if (x <= -I810_CURSOR_X) x = -I810_CURSOR_X + 1;
+   if (y <= -I810_CURSOR_Y) y = -I810_CURSOR_Y + 1;
+
+#if 0
    /*
     * There is a screen display problem when the cursor position is set
     * wholely outside of the viewport.  We trap that here, turning the
@@ -283,6 +543,7 @@ I830SetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
       show = TRUE;
       outsideViewport = FALSE;
    }
+#endif
 
    if (x < 0) {
       temp |= (CURSOR_POS_SIGN << CURSOR_X_SHIFT);
@@ -308,16 +569,30 @@ I830SetCursorPosition(ScrnInfoPtr pScrn, int x, int y)
    }
 
    /* have to upload the base for the new position */
-   if (IS_I915G(pI830) || IS_I915GM(pI830) || IS_I945G(pI830)) {
-      if (pI830->CursorIsARGB)
-         OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
-      else
-         OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
-      if (pI830->Clone) {
+   if (IS_I9XX(pI830)) {
+      if (pI830->CursorNeedsPhysical) {
          if (pI830->CursorIsARGB)
-            OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
+            OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
          else
-            OUTREG(CURSOR_B_BASE, pI830->CursorMem->Physical);
+            OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
+      } else {
+         if (pI830->CursorIsARGB)
+            OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Start);
+         else
+            OUTREG(CURSOR_A_BASE, pI830->CursorMem->Start);
+      }
+      if (pI830->Clone) {
+         if (pI830->CursorNeedsPhysical) {
+            if (pI830->CursorIsARGB)
+               OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
+            else
+               OUTREG(CURSOR_B_BASE, pI830->CursorMem->Physical);
+	 } else {
+            if (pI830->CursorIsARGB)
+               OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Start);
+            else
+               OUTREG(CURSOR_B_BASE, pI830->CursorMem->Start);
+	 }
       }
    }
 }
@@ -339,32 +614,46 @@ I830ShowCursor(ScrnInfoPtr pScrn)
 	   pI830->CursorMemARGB->Physical, pI830->CursorMemARGB->Start);
 
    pI830->cursorOn = TRUE;
-   if (IS_MOBILE(pI830) || IS_I915G(pI830) || IS_I945G(pI830)) {
+   if (IS_MOBILE(pI830) || IS_I9XX(pI830)) {
       temp = INREG(CURSOR_A_CONTROL);
-      temp &= ~(CURSOR_MODE | MCURSOR_PIPE_SELECT);
+      temp &= ~(CURSOR_MODE | MCURSOR_PIPE_SELECT | MCURSOR_GAMMA_ENABLE);
       if (pI830->CursorIsARGB)
-         temp |= CURSOR_MODE_64_ARGB_AX;
+         temp |= CURSOR_MODE_64_ARGB_AX | MCURSOR_GAMMA_ENABLE;
       else
          temp |= CURSOR_MODE_64_4C_AX;
       temp |= (pI830->pipe << 28); /* Connect to correct pipe */
       /* Need to set mode, then address. */
       OUTREG(CURSOR_A_CONTROL, temp);
-      if (pI830->CursorIsARGB)
-         OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
-      else
-         OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
-      if (pI830->Clone) {
+      if (pI830->CursorNeedsPhysical) {
+         if (pI830->CursorIsARGB)
+            OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
+         else
+            OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
+      } else {
+         if (pI830->CursorIsARGB)
+            OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Start);
+         else
+            OUTREG(CURSOR_A_BASE, pI830->CursorMem->Start);
+      }
+      if (pI830->Clone || pI830->MergedFB) {
          temp &= ~MCURSOR_PIPE_SELECT;
          temp |= (!pI830->pipe << 28);
          OUTREG(CURSOR_B_CONTROL, temp);
-         if (pI830->CursorIsARGB)
-            OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
-         else
-            OUTREG(CURSOR_B_BASE, pI830->CursorMem->Physical);
+         if (pI830->CursorNeedsPhysical) {
+            if (pI830->CursorIsARGB)
+               OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
+            else
+               OUTREG(CURSOR_B_BASE, pI830->CursorMem->Physical);
+	 } else {
+            if (pI830->CursorIsARGB)
+               OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Start);
+            else
+               OUTREG(CURSOR_B_BASE, pI830->CursorMem->Start);
+	 }
       }
    } else {
       temp = INREG(CURSOR_CONTROL);
-      temp &= ~(CURSOR_FORMAT_MASK);
+      temp &= ~(CURSOR_FORMAT_MASK | CURSOR_GAMMA_ENABLE);
       temp |= CURSOR_ENABLE;
       if (pI830->CursorIsARGB)
          temp |= CURSOR_FORMAT_ARGB;
@@ -387,7 +676,7 @@ I830HideCursor(ScrnInfoPtr pScrn)
    DPRINTF(PFX, "I830HideCursor\n");
 
    pI830->cursorOn = FALSE;
-   if (IS_MOBILE(pI830) || IS_I915G(pI830) || IS_I945G(pI830)) {
+   if (IS_MOBILE(pI830) || IS_I9XX(pI830)) {
       temp = INREG(CURSOR_A_CONTROL);
       temp &= ~CURSOR_MODE;
       temp |= CURSOR_MODE_DISABLE;
@@ -397,7 +686,7 @@ I830HideCursor(ScrnInfoPtr pScrn)
          OUTREG(CURSOR_A_BASE, pI830->CursorMemARGB->Physical);
       else
          OUTREG(CURSOR_A_BASE, pI830->CursorMem->Physical);
-      if (pI830->Clone) {
+      if (pI830->Clone || pI830->MergedFB) {
          OUTREG(CURSOR_B_CONTROL, temp);
          if (pI830->CursorIsARGB)
             OUTREG(CURSOR_B_BASE, pI830->CursorMemARGB->Physical);
@@ -428,7 +717,7 @@ I830SetCursorColors(ScrnInfoPtr pScrn, int bg, int fg)
    OUTREG(CURSOR_A_PALETTE1, fg & 0x00ffffff);
    OUTREG(CURSOR_A_PALETTE2, fg & 0x00ffffff);
    OUTREG(CURSOR_A_PALETTE3, bg & 0x00ffffff);
-   if (pI830->Clone) {
+   if (pI830->Clone || pI830->MergedFB) {
       OUTREG(CURSOR_B_PALETTE0, bg & 0x00ffffff);
       OUTREG(CURSOR_B_PALETTE1, fg & 0x00ffffff);
       OUTREG(CURSOR_B_PALETTE2, fg & 0x00ffffff);
